# copyright: Huynh Duong Phuong Vi
# www.sketchupforyou.com
module S4U
	module S4u_paint
		def self.select_faces(element,edit_transform=axes_edit_transform.inverse)
			model = Sketchup.active_model
			entities=element.definition.entities
			faces=[]
			entities.each{|e|
				if e.is_a?Sketchup::Face
					faces<< [e,edit_transform*element.transformation,element]
				elsif e.is_a?(Sketchup::Group) || e.is_a?(Sketchup::ComponentInstance)
					# e=e.make_unique if e.is_a?Sketchup::Group
					faces+=select_faces e,(edit_transform*element.transformation)
				end
			}
			return faces
		end
		def self.select_faces_material(element,material,edit_transform=axes_edit_transform.inverse)
			entities=element.definition.entities.to_a
			faces=[]
			entities.each{|e|
				if e.is_a?Sketchup::Face
					m= e.material ? e.material : material[0]
					m1= e.back_material ? e.back_material : material[0]
					faces<< [e,[m,m1],edit_transform*element.transformation,element]
				elsif e.is_a?(Sketchup::Group) || e.is_a?(Sketchup::ComponentInstance)
					# e=e.make_unique if e.is_a?Sketchup::Group
					m= e.material ? e.material : material[0]
					faces+=select_faces_material(e,[m,m],(edit_transform*element.transformation))
				end
			}
			return faces
		end
		def self.axes_edit_transform
			model=Sketchup.active_model
			return ( Sketchup.version.to_i>=16) ? model.axes.transformation : model.edit_transform
		end
		def self.select_elements(element)
			model = Sketchup.active_model
			entities=element.definition.entities
			elements=[]
			entities.each{|e|
				if e.is_a?(Sketchup::Group) || e.is_a?(Sketchup::ComponentInstance)
					# e=e.make_unique if e.is_a?Sketchup::Group
					elements<< e
					elements+=select_elements(e)
				end
			}
			return elements
		end
		def self.get_min_max(ps)
			p1,p2=ps[0],ps[-1]
			if ps[0].is_a?Sketchup::Vertex
				ps1=ps.sort_by{|p| p.position.z}
				ps1=ps1.select{|p| p.position.z==ps1[0].position.z}
				ps1=ps1.sort_by{|p| p.position.y}
				ps1=ps1.select{|p| p.position.y==ps1[0].position.y}
				p1=ps1.sort_by{|p| p.position.x}[0]

				ps1=ps.sort_by{|p| p.position.z}
				ps1=ps1.select{|p| p.position.z==ps1[-1].position.z}
				ps1=ps1.sort_by{|p| p.position.y}
				ps1=ps1.select{|p| p.position.y==ps1[-1].position.y}
				p2=ps1.sort_by{|p| p.position.x}[-1]
			else
				ps1=ps.sort_by{|p| p.z}
				ps1=ps1.select{|p| p.z==ps1[0].z}
				ps1=ps1.sort_by{|p| p.y}
				ps1=ps1.select{|p| p.y==ps1[0].y}
				p1=ps1.sort_by{|p| p.x}[0]
				ps1=ps.sort_by{|p| p.z}
				ps1=ps1.select{|p| p.z==ps1[-1].z}
				ps1=ps1.sort_by{|p| p.y}
				ps1=ps1.select{|p| p.y==ps1[-1].y}
				p2=ps1.sort_by{|p| p.x}[-1]

			end
			return [p1,p2]
			# rescue
			# p "error"
			# p1,p2=ps[0],ps[-1]
			# if ps[0].is_a?Sketchup::Vertex
				# ps=ps.sort_by{|p| p.position.distance(ORIGIN)}
				# p1,p2=ps[0],ps[-1]
			# else
				# ps=ps.sort_by{|p| p.distance(ORIGIN)}
				# p1,p2=ps[0],ps[-1]
			# end
			# return [p1,p2]
		end
		def self.get_area_material(edit_material,material)
			model=Sketchup.active_model
			sel=model.selection.to_a
			sel=model.active_entities.to_a if sel.empty?
			objects=sel.select{|e|(e.is_a?Sketchup::Group) || (e.is_a?Sketchup::ComponentInstance) ||
							(e.is_a?Sketchup::Face)}
			if objects.length>0
				edit_material1 = material ? material : edit_material
				selects=[]
				faces=[]
				faces1=[]
				areas=[]
				area=0
				objects.each{|e|
					if (e.is_a?Sketchup::Group) || (e.is_a?Sketchup::ComponentInstance)
						faces += self.select_faces_material(e,[edit_material,edit_material])
					elsif (e.is_a?Sketchup::Face)
						selects<<e
					end
				}
				area1=0
				if faces && faces.length>0
					faces.each{|f|
						if f[1][0]==edit_material1
							s= f[0].area(f[2])
							areas << [f[0],s]
						elsif f[1][1]==edit_material1
							s= f[0].area(f[2])
							areas<< [f[0],s]
						end
					}
				end
				selects.each{|f|
						if f.material==material
							areas<<[f,f.area]
						elsif (f.back_material==material)
							areas<<[f,f.area]
						end
				}
				areas.each{|f| area+= f[1]}
				UI.messagebox (Strings["Total Area"] +": "+  areas.length.to_s  + " "+ Strings["Faces"] + "\n" +
				Sketchup.format_area(area).to_s),MB_MULTILINE,Strings["Area Of Material"]
			else
				UI.messagebox (Strings["Select Material Face"])
			end
		end
		def self.set_materials(objects,material=nil,back_material=nil)
			objects.uniq!
			selects=[]
			faces=[]
			objects.each{|e|
				if (e.is_a?Sketchup::Group) || (e.is_a?Sketchup::ComponentInstance)
					faces+=S4U::S4u_paint.select_faces(e)
					selects+= select_elements(e)
				elsif e.is_a?Sketchup::Drawingelement
					selects<<e
				end
			}
			selects+=faces.transpose[0] if faces && faces.length>0
			selects.each{|f|
					f.material = material
					f.back_material = back_material if f.is_a?Sketchup::Face
			}
			objects.each{|e| e.material=material}
		end
		def self.get_scale(transformation)
			a=transformation.to_a
			sx=Math.sqrt(a[0]**2+a[1]**2+a[2]**2) / a[15]
			sy=Math.sqrt(a[4]**2+a[5]**2+a[6]**2) / a[15]
			sz=Math.sqrt(a[8]**2+a[9]**2+a[10]**2) / a[15]
			return [sx,sy,sz]
		end
		def self.normal_face(face)
			ver=face.outer_loop.vertices.collect{|e| e.position}
			vx=nil;vy=nil;line=nil
			(0..ver.length-2).each{|i|
				vx=ver[i].vector_to ver[i+1]
				if vx.length!=0
					line=[ver[i],ver[i+1]]
					(0..ver.length-1).each{|j|
					v2=ver[j].project_to_line line
					vy=v2.vector_to ver[j]
					if vy.length!=0
						vz = (vx.cross vy)
						vz.reverse! if vz.reverse.angle_between(face.normal)<Math::PI*0.5
						return vz
					end
					}
				end
			}
			return face.normal
		end
		def self.reverse_face_view(sels)
			model=Sketchup.active_model
			view=model.active_view
			faces=[]
			selects=[]
			sels.each{|e|
					if (e.is_a?Sketchup::Group) || (e.is_a?Sketchup::ComponentInstance)
						faces+=S4U::S4u_paint.select_faces(e)
					else
						selects<< e if e.is_a?Sketchup::Face
					end
			}
			face_reverse=[]
			i=0
			length=faces.length+selects.length
			faces.each{|fs|
				i+=1
				Sketchup.status_text= ("Processing: " + i.to_s + "/" + length.to_s)
				model.active_view.invalidate
				f=fs[0]
				t=model.edit_transform*fs[1]
				entities=model.active_entities.to_a
				normal=S4U::S4u_paint.normal_face(f).transform(t)
				vertices=S4U::S4u_paint.get_vertices(f).collect{|e| e.position.transform(t)}
				flag=true;ray=nil
				vertices.each{|v|
					pv=view.screen_coords(v)
					ray = view.pickray pv.x,pv.y
					ray1=[v,ray[1].reverse]
					e = model.raytest(ray1,true)
					if e
						(flag=false;break) if (e[1] & entities).length>0
					end
				}
				if flag
					vec= ray[1].reverse.angle_between(normal.reverse) < Math::PI*0.5
					face_reverse<<f if vec
				end
			}

			selects.each{|fs|
				i+=1
				Sketchup.status_text= ("Processing: " + i.to_s + "/" + length.to_s)
				model.active_view.invalidate
				f=fs
				t=model.edit_transform
				entities=model.active_entities.to_a
				vertices=S4U::S4u_paint.get_vertices(f).collect{|e| e.position}
				normal=f.normal
				flag=true;ray=nil
				vertices.each{|v|
					pv=view.screen_coords(v)
					ray = view.pickray pv.x,pv.y
					ray1=[v,ray[1].reverse]
					e = model.raytest(ray1,true)
					if e
						(flag=false;break) if (e[1] & entities.to_a).length>0
					end
				}
				if flag
					vec= ray[1].reverse.angle_between(normal.reverse) < Math::PI*0.5
					face_reverse<< f if vec
				end
			}
			face_reverse.uniq.each{|e| e1= e.reverse! }
		end
		def self.find_connected(fa)
			faces=[fa]
			all_connected=fa.all_connected.grep(Sketchup::Face).select{|f|
			(f.loops.collect{|l| l.edgeuses.collect{|e| e.edge}}.flatten).any?{|e| e.soft?}}
			i=0
			all_connected=all_connected-faces
			length=all_connected.length
			# t=Time.now.to_f
			faces1=[fa]
			eds=[]
			while all_connected.length>0 && i<length
				faces2=[]
				eds1=[]
				faces1.each{|f|
					ed=f.loops.collect{|l| l.edgeuses.collect{|e| e.edge}}.flatten.select{|e| e.soft?}.uniq
					ed=ed - eds
					eds1 += ed
					faces2 += ed.collect{|e| e.faces}
				}
				faces2=faces2.flatten.uniq
				faces2= faces2-faces
				break if faces2.empty?
				faces1=faces2.clone
				eds=eds1.clone
				faces+=faces1
				all_connected=all_connected-faces1
				i+=1
			end
			# p (Time.now.to_f-t),"asdf"
			return faces
		end
		def self.find_surface(fa)
			model=Sketchup.active_model
			faces_soft=[[[fa],nil]]
			faces=[fa]
			all_connected=fa.all_connected.grep(Sketchup::Face).select{|f|
			(f.loops.collect{|l| l.edgeuses.collect{|e| e.edge}}.flatten).any?{|e| e.soft?}}
			i=0
			all_connected=all_connected-faces
			length=all_connected.length
			# t=Time.now.to_f
			faces1=[fa]
			eds=[]
			while all_connected.length>0 && i<length
				faces2=[]
				eds1=[]
				faces1.each{|f|
					ed=f.loops.collect{|l| l.edgeuses.collect{|e| e.edge}}.flatten.select{|e| e.soft?}.uniq
					ed=ed - eds
					eds1 += ed
					fs= ed.collect{|e| e.faces}
					fs=fs.flatten.uniq
					fs= fs-faces
					if !fs.empty?
						faces_soft<<[fs,f]
						faces2+=fs
					end
				}
				faces2=faces2.flatten.uniq
				faces2= faces2-faces
				break if faces2.empty?
				faces1=faces2.clone
				eds=eds1.clone
				faces+=faces1
				all_connected=all_connected-faces1
				i+=1
			end
			# p (Time.now.to_f-t),"asdf"
			return faces_soft,faces.uniq
		end
		def self.copy(f,trans)
			origin,xaxis,yaxis,zaxis=trans.origin.clone,trans.xaxis.clone,trans.yaxis.clone,trans.zaxis.clone
			plane=[f.origin,f.zaxis]
			plane1=[trans.origin,trans.zaxis]
			if trans.zaxis.parallel? f.zaxis
				if origin.on_plane?(plane)
					origin=f.origin
					xaxis,yaxis=f.xaxis,f.yaxis
				else
					xaxis,yaxis=f.xaxis,f.yaxis
					origin=f.origin.project_to_plane(plane1)
				end
			else
				line=Geom::intersect_plane_plane(plane,plane1)
				if line
					angle=f.zaxis.angle_between trans.zaxis
					t=Geom::Transformation.rotation line[0],line[1],angle
					origin=f.origin.transform t
					xaxis=f.xaxis.transform t
					yaxis=f.yaxis.transform t
				end
			end
			t=Geom::Transformation.new(xaxis,yaxis,zaxis,origin)
			return t
		end
		def self.get_vertices(face)
			ver=[]
			vertices=[]
			vertices =face.outer_loop.vertices*3
			(0..face.outer_loop.vertices.length-1).each { |i|
				j=face.outer_loop.vertices.length+i
				p1=vertices[j-1].position.clone
				p2=vertices[j].position.clone
				p3=vertices[j+1].position.clone
				v1=p1.vector_to p2
				v2=p2.vector_to p3
				ver.push vertices[j] if not( v1.samedirection? v2)
			}
			return ver
		end
		def self.get_info_texture(face,trans,front=true)
			x,y =1,1
			pts1 = []
			pts1[0] = Geom::Point3d.new [0, 0, 0]
			pts1[1] = Geom::Point3d.new [x, 0, 0]
			pts1[2] = Geom::Point3d.new [x, y, 0]
			pts1[3] = Geom::Point3d.new [0, y, 0]
			trans_face=Geom::Transformation.new ORIGIN,(front ? face.normal : face.normal.reverse)
			material=front ? face.material : face.back_material
			pts1=pts1.collect{|p| p.transform trans_face}
			uvHelp=face.get_UVHelper(front,front)
			if front
				points=pts1.collect{|p| uvHelp.get_front_UVQ(p)}
			else
				points=pts1.collect{|p| uvHelp.get_back_UVQ(p)}
			end
			s1=points[1].y-points[0].y
			ang=nil
			if s1!=0
				s1= (points[2].y-points[0].y)/s1
				s1= s1==1 ? 1 : 1/(1-s1)
				s2=points[3].y-points[0].y
				if s2!=0
					s2=(points[2].y-points[0].y)/s2
					s2= s2==1 ? 1 : 1/(1-s2)
				else
					ang = Math::PI*0.5
				end
			else
				ang=0
			end
			if !ang
				p1=Geom::Point3d.new(1,s1,0)
				p2=Geom::Point3d.new(s2,1,0)
				vec=p2-p1
				if vec.length==0
					ang=Math::PI*0.25
				else
					ang=vec.angle_between(X_AXIS.transform(trans_face.inverse))
				end
			end
			sy=(points[1].y-points[0].y)
			if sy.abs<=0.001
				sy= (points[3].y-points[0].y)
				sy=1/sy
			else
				sy= Math::sin(ang)/sy
			end
			sx= (points[1].x-points[0].x)

			if sx.abs<=0.001
				sx= (points[3].x-points[0].x)
				sx= 1/sx
			else
				sx= Math::cos(ang)/sx
			end
			mx,my = material.texture.width,material.texture.height
			return [ang,(sx/mx).abs,(sy/my).abs]
		end
		class Paint
			def initialize(setting)
				i=-1
				@paint_mode=setting[i+=1]
				@aperture=setting[i+=1]
				@angle=setting[i+=1]
				@make_unique=setting[i+=1]

				@cursor_id={}
				ext=IS_OSX ? ".pdf" : ".svg"
				cursor_path= File.join(File.dirname(__FILE__),NAME)  + ext
				@cursor_id["paint"] = UI.create_cursor(cursor_path, 15, 0)
				cursor_path= File.join(File.dirname(__FILE__),NAME+"_multiple") + ext
				@cursor_id["multiple"] = UI.create_cursor(cursor_path, 15, 0)
				cursor_path= File.join(File.dirname(__FILE__),NAME+"_connected")+ ext
				@cursor_id["connected"] = UI.create_cursor(cursor_path, 15, 0)
				cursor_path= File.join(File.dirname(__FILE__),NAME+"_delete")+ ext
				@cursor_id["delete"] = UI.create_cursor(cursor_path, 15, 0)
				cursor_path= File.join(File.dirname(__FILE__),NAME+"_replace")+ ext
				@cursor_id["replace"] = UI.create_cursor(cursor_path, 15, 0)
				cursor_path= File.join(File.dirname(__FILE__),NAME+"_sample")+ ext
				@cursor_id["sample"] = UI.create_cursor(cursor_path, 15, 24)
				cursor_path= File.join(File.dirname(__FILE__),NAME+"_object")+ ext
				@cursor_id["object"] = UI.create_cursor(cursor_path, 15, 0)
				cursor_path= File.join(File.dirname(__FILE__),NAME+"_reverse")+ ext
				@cursor_id["reverse"] = UI.create_cursor(cursor_path, 15, 0)
				cursor_path= File.join(File.dirname(__FILE__),NAME+"_align_mouse")+ ext
				@cursor_id["align"] = UI.create_cursor(cursor_path, 0 , 0)
			end
			def onSetCursor
				#if @paint_mode==0  || @paint_mode==1
					if @paint_mode!=4
						if @alt_press
							UI.set_cursor(@cursor_id["sample"])
						elsif @shift_press && @paint_mode!=3
							UI.set_cursor(@cursor_id["multiple"])
						elsif @ctrl_press && @paint_mode!=3
							UI.set_cursor(@cursor_id["connected"])
						elsif @paint_mode==1
							UI.set_cursor(@cursor_id["object"])
						elsif @paint_mode==2
							UI.set_cursor(@cursor_id["delete"])
						elsif @paint_mode==3
							UI.set_cursor(@cursor_id["replace"])
						# elsif @paint_mode==4
							# UI.set_cursor(@cursor_id["align"])
						elsif @paint_mode==5
							UI.set_cursor(@cursor_id["reverse"])
						else
							UI.set_cursor(@cursor_id["paint"])
						end
					else
						UI.set_cursor(@cursor_id["align"])
					end
				#end
			end
			def activate
				@strings=S4U::S4u_paint::Strings
				@ip=Sketchup::InputPoint.new
				@point_start=Sketchup::InputPoint.new
				@point_end=Sketchup::InputPoint.new
				@origin=Sketchup::InputPoint.new
				@make_unique=false
				@paint_list=[@strings.GetString("Paint Face"),@strings.GetString("Paint Object"),
							@strings.GetString("Delete Material"),@strings.GetString("Replace Material"),
							@strings.GetString("Align Move Rotate"),
							@strings.GetString("Reverse Face"),@strings.GetString("Area Of Material"),
							@strings.GetString("Align Material To Line"),@strings.GetString("Move Material"),
							@strings.GetString("Make Unique")
				]
				@picked_elements={}
				@picked_elements_move={}
				@edit_material=nil
				@select_mode=0
				Sketchup.active_model.active_path.reverse.each{|e| (@edit_material = e.material;break) if e.material} if Sketchup.active_model.active_path
				self.reset(nil)
				self.setstatus(nil)
			end
			def deactivate(view)
				(@model_active.commit_operation ; @model_active=nil) if @model_active
				setting=[@paint_mode,@aperture,@angle,@make_unique]
				S4U::S4u_paint.instance_variable_set :@setting,setting
			end
			def draw(view)
				if @paint_mode==4
					if @select_mode==1 || (@alt_press && @select_mode==0)
						draw_edge_over(view) if !@picked_edge
					end
					if (@select_mode==2 || (@select_mode==0 && @shift_press)) && @flags
						if @ip && @ip.valid? && @ip.display?
							@ip.draw(view)
							view.draw_points @ip.position,10,2,"red"
						end
					end
					@ip.draw(view) if @ip && @ip.valid? && @ctrl_press && @select_mode==0
					if @picked_edge && @select_mode!=2
						draw_edge_select(view,@picked_edge)
					elsif @select_mode==0 && @origin && @origin.valid? && !@picked_edge
						@origin.draw(view)
						view.draw_points @origin.position,20,3,"red"
						view.draw_points @origin.position,20,1,"red"
					end
				else
					if @ip && @ip.valid? && @ip.display? && @flags
						aperture=@aperture>10 ? @aperture : 10
						view.draw_points @ip.position,aperture,1
					end
				end
			end
			def draw_edge_over(view)
				view.drawing_color = "blue"
				view.line_width = 4
				if @selectedge
					view.set_color_from_line(@selectedge[0],@selectedge[1]) if @select_axis
					view.draw_line(@selectedge[0],@selectedge[1])
				end
			end
			def draw_edge_select(view,picked_edge)
				view.line_width = 5
				if picked_edge
					view.drawing_color = "blue"
					view.draw_line(picked_edge[0],picked_edge[1])
				end
			end
			def getMenu(menu)
				list=[]
				@paint_list[0..4].each_index{|i|
					list<<menu.add_item(@paint_list[i]) {
						@paint_mode=i
						 @select_mode=0 if i==4
						self.setstatus
					}
					menu.set_validation_proc(list[-1]){ MF_UNCHECKED}
				}
				menu.add_separator
				if @paint_mode==4
					list1=[]
					list1<<menu.add_item(@paint_list[7]) {
						@select_mode= @select_mode==1 ? 0 : 1
						self.setstatus
					}
					menu.set_validation_proc(list1[-1]){ MF_UNCHECKED}
					menu.set_validation_proc(list1[-1]){ MF_CHECKED} if @select_mode==1
					list1<<menu.add_item(@paint_list[8]) {
						@select_mode= @select_mode==2 ? 0 : 2
						self.setstatus
					}
					menu.set_validation_proc(list1[-1]){ MF_UNCHECKED}
					menu.set_validation_proc(list1[-1]){ MF_CHECKED} if @select_mode==2
					menu.add_separator
				end
				@paint_list[5..6].each_index{|i|
					list<<menu.add_item(@paint_list[i+5]) { @paint_mode=i+5 ;self.setstatus}
					menu.set_validation_proc(list[-1]){ MF_UNCHECKED}
				}
				menu.set_validation_proc(list[@paint_mode]){ MF_CHECKED}
				list<<menu.add_item(@paint_list[9]) { @make_unique=!@make_unique;self.setstatus}
				menu.set_validation_proc(list[-1]){ @make_unique ?  MF_CHECKED : MF_UNCHECKED }
			end
			def get_picked_material(flags,view,x,y,face,transformation)
				ray = view.pickray x, y
				# normal=face.normal.transform transformation
				normal=S4U::S4u_paint.normal_face(face).transform transformation
				vec= ray[1].reverse.angle_between(normal) < Math::PI*0.5
				material= vec ? face.material : face.back_material
				return [material,vec]
			end
			def get_material_out(path)
				material=nil
				path.reverse[1..-1].each{|e|
					(material=e.material;break) if e.material
				} if path && path.length>1
				return material
			end
			def get_bounds(element)
				a=element.transformation.to_a
				b=nil
				if element.is_a?Sketchup::ComponentInstance
					b=element.definition.bounds
				elsif element.is_a?Sketchup::Group
					b=element.local_bounds
				else
					return nil
				end
				return (0..7).collect{|i| b.corner(i).clone.transform! element.transformation } if b
			end
			def select_boundingbox(ip,x,y,view)
				model=Sketchup.active_model
				entities=model.active_entities
				position=Geom::Point3d.new(x,y,0)
				#view=model.active_view
				gc=entities.select{|e| ((e.is_a?Sketchup::Group) || (e.is_a?Sketchup::ComponentInstance)) &&
										(view.screen_coords(e.bounds.center).distance(position) <
										view.screen_coords(e.bounds.center).distance(view.screen_coords(e.bounds.max))*1.5) &&
										e.layer.visible?
										#e.bounds.center.distance(ip.position) < e.bounds.diagonal*0.7
				}.sort_by{|e| e.bounds.center.distance(ip.position) }
				gc[0..7]
				edges=[]
				gc.each{|e|
					bounds=get_bounds(e)
					eds=[
					[bounds[0],bounds[1]],[bounds[0],bounds[2]],[bounds[0],bounds[4]],
					[bounds[1],bounds[3]],[bounds[1],bounds[5]],
					[bounds[2],bounds[3]],[bounds[2],bounds[6]],
					[bounds[3],bounds[7]],
					[bounds[4],bounds[5]],[bounds[4],bounds[6]],
					[bounds[5],bounds[7]],[bounds[6],bounds[7]],
					[Geom.linear_combination(0.5,bounds[0],0.5,bounds[1]),Geom.linear_combination(0.5,bounds[2],0.5,bounds[3])],
					[Geom.linear_combination(0.5,bounds[0],0.5,bounds[1]),Geom.linear_combination(0.5,bounds[4],0.5,bounds[5])],
					[Geom.linear_combination(0.5,bounds[4],0.5,bounds[5]),Geom.linear_combination(0.5,bounds[6],0.5,bounds[7])],
					[Geom.linear_combination(0.5,bounds[6],0.5,bounds[7]),Geom.linear_combination(0.5,bounds[2],0.5,bounds[3])],

					[Geom.linear_combination(0.5,bounds[0],0.5,bounds[2]),Geom.linear_combination(0.5,bounds[1],0.5,bounds[3])],
					[Geom.linear_combination(0.5,bounds[1],0.5,bounds[3]),Geom.linear_combination(0.5,bounds[5],0.5,bounds[7])],
					[Geom.linear_combination(0.5,bounds[5],0.5,bounds[7]),Geom.linear_combination(0.5,bounds[4],0.5,bounds[6])],
					[Geom.linear_combination(0.5,bounds[4],0.5,bounds[6]),Geom.linear_combination(0.5,bounds[0],0.5,bounds[2])],

					[Geom.linear_combination(0.5,bounds[0],0.5,bounds[4]),Geom.linear_combination(0.5,bounds[1],0.5,bounds[5])],
					[Geom.linear_combination(0.5,bounds[1],0.5,bounds[5]),Geom.linear_combination(0.5,bounds[3],0.5,bounds[7])],
					[Geom.linear_combination(0.5,bounds[3],0.5,bounds[7]),Geom.linear_combination(0.5,bounds[2],0.5,bounds[6])],
					[Geom.linear_combination(0.5,bounds[2],0.5,bounds[6]),Geom.linear_combination(0.5,bounds[0],0.5,bounds[4])]
					]
					eds.select!{|ed| ed[0].distance(ed[1])>0}
					eds.map!{|ed| [ed,e] }
					edges+=eds
				}
				edges.sort_by!{|e| view.screen_coords(ip.position).distance_to_line([view.screen_coords(e[0][0]),view.screen_coords(e[0][1])]) }
				e=edges[0]
				return e if e && (view.screen_coords(ip.position).distance_to_line([view.screen_coords(e[0][0]),view.screen_coords(e[0][1])]) < 7)
				return nil
			end
			def onMouseMove(flags, x, y, view)
				@shift_press= (CONSTRAIN_MODIFIER_MASK&flags==CONSTRAIN_MODIFIER_MASK)
				@ctrl_press= (COPY_MODIFIER_MASK&flags==COPY_MODIFIER_MASK)
				@alt_press= (ALT_MODIFIER_MASK&flags==ALT_MODIFIER_MASK)
				@flags=((flags&1)==1)
				@ip.pick view,x,y
				@ph_select = view.pick_helper
				@ph_select.do_pick(x, y,@aperture)

				if @paint_mode==4 && (@select_mode==2 || (@select_mode==0 && @shift_press))
					if (x-@x).abs>2 && (y-@y).abs>2
						@x,@y=x,y
						@point_end.copy! @ip if @ip.valid? &&  @ip.display?
						if @mats && @point_start.valid? && @point_end.valid?
							@model_active.abort_operation if @model_active
							view.model.start_operation TOOL,true,false,false
							@model_active=view.model
							move_material @mats[0],@mats[1],[@point_start.position,@point_end.position]
							paint_face_soft if !Sketchup.active_model.rendering_options["DrawHidden"]
							self.setstatus(view)
							return
						end
					end
				elsif @paint_mode==4 && ((@select_mode==0 && @ctrl_press))
					self.setstatus(view)
				elsif @paint_mode==4 && (@select_mode>0 || (@select_mode==0 && @alt_press))
					tooltip=@strings["Material:"]+ " "
					@selectedge=nil;@select_axis=nil
					selectedge=@ip.edge
					element = @ph_select.best_picked
					trans=@ip.transformation
					tedit=S4U::S4u_paint.axes_edit_transform
					ip1 = view.inputpoint(x, y,Sketchup::InputPoint.new(tedit.origin))
					p=ip1.position.transform tedit.inverse
					if @ph_select.picked_edge
						selectedge=@ph_select.picked_edge
						pstart=selectedge.start.position.clone.transform trans
						pend=selectedge.end.position.clone.transform trans
						@selectedge=[pstart,pend]
					elsif selectedge
						pstart=selectedge.start.position.clone.transform trans
						pend=selectedge.end.position.clone.transform trans
						@selectedge=[pstart,pend]
					elsif (bd=select_boundingbox(@ip,x,y,view)) && bd
						@selectedge=bd[0]
						view.tooltip=@strings.GetString("BoundingBox")
					elsif element.is_a?(Sketchup::ConstructionLine)
						vector = element.direction
						length= view.pixels_to_model 100, ORIGIN
						if vector.length>0
							vector.length=length
							pstart=@ip.position.clone + vector
							pend=@ip.position.clone + vector.reverse
							@selectedge=[pstart,pend]
						end
						view.tooltip=@strings.GetString("ConstructionLine")
					elsif ((p.x==0 && p.y==0) || (p.x==0 && p.z==0) || (p.y==0 && p.z==0))
						@select_axis=true
						vector =tedit.origin.vector_to ip1.position
						length= view.pixels_to_model 100, ORIGIN
						if vector.length>0
							vector.length=length
							pstart=ip1.position.clone + vector
							pend=ip1.position.clone + vector.reverse
							@selectedge=[pstart,pend]
						end
						view.tooltip=@strings.GetString("Axis")
					end
					self.setstatus(view) #if @flags
				elsif @paint_mode!=4
					tooltip=@strings["Material:"]+ " "
					@material=false
					@face_move=nil
					if @alt_press  && (@ip.face || @ip.edge)
						face=@ip.face
						if face
							@material,vec=get_picked_material(flags,view,x,y,face,@ip.transformation)
							@face_move=[face,vec,@ip.transformation]
						elsif @ip.edge
							@material=@ip.edge.material
						end
						if !@material
							if Sketchup.version.to_i>=17
								@material= get_material_out(@ip.instance_path.to_a)
							end
						end
						tooltip+= @material ? @material.display_name : @strings["Default"]
						self.setstatus(nil,tooltip)
					else
						if ((flags&1)==1) || @ph_select.best_picked.is_a?(Sketchup::Drawingelement)
							if @alt_press
								face=@ph_select.picked_face
								if face
									@material,vec=get_picked_material(flags,view,x,y,face,@ip.transformation)
									@face_move=[face,vec,@ip.transformation]
								elsif @ph_select.picked_edge
									@material=@ph_select.picked_edge.material
								elsif @ph_select.best_picked
									@material=@ph_select.best_picked.material
								end
								@material= get_material_out(@ph_select.path_at(0)) if !@material
								@material = @edit_material if !@material
								tooltip+= @material ? @material.display_name : @strings["Default"]
								self.setstatus(nil,tooltip)
							elsif ((flags&1)==1)
								get_picked(@ph_select,flags,view,x,y) if @ph_select
							end
						end
					end
				end
			end
			def get_picked(ph_select,flags,view,x,y)
				if @make_unique && @paint_mode!=3 && @paint_mode!=6
					view.model.start_operation TOOL,true,false,false
					ph_select.count.times { |index|
						if ph_select.leaf_at(index).is_a?(Sketchup::Drawingelement)
							path=ph_select.path_at(index)
							path.each{|e| e.make_unique if e.is_a?(Sketchup::Group) || e.is_a?(Sketchup::ComponentInstance) }
							ph_select.do_pick(x, y,@aperture)
							@ip.pick view,x,y
						end
					}
					view.model.commit_operation
				end
				ph_select.count.times { |index|
					if ph_select.leaf_at(index).is_a?(Sketchup::Drawingelement)
						vec=true
						if ph_select.leaf_at(index).is_a?Sketchup::Face
							face=ph_select.leaf_at(index)
							material,vec=get_picked_material flags,view,x,y,face,ph_select.transformation_at(index)
						end
						id=ph_select.element_at(index).entityID.to_s + "." + ph_select.leaf_at(index).entityID.to_s
						@picked_elements[id] =
						[ph_select.leaf_at(index),ph_select.element_at(index),ph_select.transformation_at(index),vec,index]
					end
				}
			end
			def select_surface(leaf,element,trans,vec,index=0)
				picked_elements={}
				faces_soft_hash={}
				if (!Sketchup.active_model.rendering_options["DrawHidden"]) && (leaf.is_a?Sketchup::Face)
					face=leaf
					faces_soft=[]
					if face.edges.any?{|e| e.soft?}
						faces,faces_list=S4U::S4u_paint.find_surface(face)
						faces.each{|fs|
							ft=[]
							fs[0].each{|f|
								id=element.entityID.to_s + "." + f.entityID.to_s
								picked_elements[id] = [f,element,trans,vec,index]
								ft<<picked_elements[id]
							}
							if fs[1]
								id=element.entityID.to_s + "." + fs[1].entityID.to_s
								faces_soft<<[ft,picked_elements[id]]
							else
								id=element.entityID.to_s + "." + face.entityID.to_s
								faces_soft<<[ft,picked_elements[id]]
							end

						}
						fs=faces_soft.collect{|f| f.flatten(1)[0]}
						id1=fs.sort{|f| f[1].entityID }[0]
						faces_soft_hash[id1]=faces_soft if faces_soft.length>0
					end
				end
				return [picked_elements,faces_soft_hash,faces_list]
			end
			def onLButtonDown(flags, x, y, view)
				get_picked(@ph_select,flags,view,x,y) if @ph_select && @make_unique
				if @paint_mode==4
					if @select_mode==2 || (@select_mode==0 && @shift_press)
						model=Sketchup.active_model
						model.start_operation TOOL,true,false,false
						@model_active=model
						@point_start.copy! @ip if @ip.valid? &&  @ip.display?
						get_picked(@ph_select,flags,view,x,y) if @ph_select
						@mats=get_faces(0,false,false)
					end
					if (@select_mode==0 && @ctrl_press && !@picked_edge)
						@origin.copy! @ip if @ip.valid? &&  @ip.display?
					end
				end
			end
			def onLButtonUp(flags, x, y, view)
				if @paint_mode!=4 #|| (@select_mode==0 && @paint_mode==4)
					if @alt_press
						@face_material=@face_move.clone if @face_move
						Sketchup.active_model.materials.current= @material if @material!=false
						@picked_elements.clear
						self.setstatus(view)
					else
						get_picked(@ph_select,flags,view,x,y) if @ph_select
						if @picked_elements.length>0
							process
							self.reset(view)
						end
						self.setstatus(view)
					end
				else
					if @select_mode==2 || (@select_mode==0 && @shift_press)
						if @ip.valid? && @ip.display?
							@point_end.copy! @ip
							if @mats && @point_start.valid? && @point_end.valid?
								@model_active.abort_operation if @model_active
								view.model.start_operation TOOL,true,false,false
								@model_active=view.model
								move_material @mats[0],@mats[1],[@point_start.position,@point_end.position]
								paint_face_soft if !Sketchup.active_model.rendering_options["DrawHidden"]
							end
							@mats=nil
							(@model_active.commit_operation ; @model_active=nil) if @model_active
							@point_start.clear
							@point_end.clear
							self.reset(view)
							self.setstatus(view)
						end
					elsif @select_mode!=2  && !@ctrl_press
						if (@alt_press || @select_mode==1) && @selectedge && !@picked_edge
							@origin.clear
							@picked_edge=@selectedge
						else
							get_picked(@ph_select,flags,view,x,y) if @ph_select
							if @picked_elements.length>0 && !@alt_press
								process
								# @origin.clear
								self.reset(view)
								self.setstatus(view)
							end
						end
					end
				end
			end
			def process()
				model=Sketchup.active_model
				model.start_operation TOOL,true,false,false
				if (@paint_mode==0 || @paint_mode==1)
					mats=get_faces @paint_mode,@shift_press,@ctrl_press
					paint_material mats[0],mats[1]
					paint_face_soft(@face_material) if !Sketchup.active_model.rendering_options["DrawHidden"]
				elsif @paint_mode==2
					delete_material
				elsif @paint_mode==3
					replace_material
				elsif @paint_mode==4
					mats=get_faces(0,false,false)
					if @select_mode==1 || (@select_mode==0 && @picked_edge)
						align_material mats[0],mats[1]
					elsif @select_mode==2 || (@select_mode==0 && @shift_press)
						move_material mats[0],mats[1],[@point_start.position,@point_end.position]
					elsif @select_mode==0
						rotate_material mats[0],mats[1]
					end
					paint_face_soft if !Sketchup.active_model.rendering_options["DrawHidden"]
				elsif @paint_mode==5
					reverse_face
				elsif @paint_mode==6
					id=@ph_select.best_picked.entityID.to_s + "." + @ph_select.picked_face.entityID.to_s
					material=get_material(id)
					S4U::S4u_paint.get_area_material(@edit_material,material)
				end
				model.commit_operation
				rescue => error
				p error
				p error.backtrace.join("\n")
				model.abort_operation
			end
			def reverse_face
				sels=Sketchup.active_model.selection.to_a
				back_materials=[]
				materials=[]
				selects=@picked_elements.each_value.collect{|e| [e[0],e[3],e[1]] }.compact
				if sels.empty?
				#if !selects.empty?
					if @shift_press
						selects.each{|e|
							id =e[2].entityID.to_s + "." + e[0].entityID.to_s
							material=@picked_elements[id][3] ? @picked_elements[id][0].material : @picked_elements[id][0].back_material
							sel=[]
							if e[2].is_a?(Sketchup::Group) || e[2].is_a?(Sketchup::ComponentInstance)
								sel=e[0].parent.entities.to_a #if sels.include?(e[2])
							else
								sel=(e[0].parent.entities.to_a) #& sels
							end
							sel.each{|e1|
								if e1.is_a?Sketchup::Drawingelement
									back_materials << e1 if e1.is_a?(Sketchup::Face) && (e1.back_material==material)
									materials<< e1 if (e1.material==material)
								end
							}
						}
					elsif @ctrl_press
						selects.each{|e|
							id =e[2].entityID.to_s + "." + e[0].entityID.to_s
							material=@picked_elements[id][3] ? @picked_elements[id][0].material : @picked_elements[id][0].back_material
							faces=(e[0].all_connected.to_a).grep(Sketchup::Drawingelement)
							sel=[]
							if e[2].is_a?(Sketchup::Group) || e[2].is_a?(Sketchup::ComponentInstance)
								sel=faces# if sels.include?(e[2])
							else
								sel=faces #& sels
							end
							sel.each{|e1|
								if e1.is_a?Sketchup::Drawingelement
									back_materials << e1 if e1.is_a?(Sketchup::Face) && (e1.back_material==material)
									materials<< e1 if (e1.material==material)
								end
							}
						}
					else
						selects.each{|e|
							if e[0].is_a?(Sketchup::Face) && !e[1]
								back_materials<< e[0]
							else
								materials << e[0]
							end
						}
					end
					mats=[]
					# mats=get_face_soft.flatten! if !Sketchup.active_model.rendering_options["DrawHidden"]
					(materials+back_materials+mats).uniq.grep(Sketchup::Face).each{|f| f.reverse! }
				else
					S4U::S4u_paint.reverse_face_view(sels) if !sels.empty?
				end
			end
			def get_faces(paint_mode=0,shift_press=true,ctrl_press=true)
				sels=Sketchup.active_model.selection.to_a
				sels=Sketchup.active_model.active_entities.to_a if sels.empty?
				back_materials=[]
				materials=[]
				if paint_mode==0
					selects=@picked_elements.each_value.collect{|e| [e[0],e[1],e[2],e[3]]} #[e[0],e[3],e[1],e[2]]}
					if shift_press
						selects.each{|e|
							id =e[1].entityID.to_s + "." + e[0].entityID.to_s
							material=@picked_elements[id][3] ? @picked_elements[id][0].material : @picked_elements[id][0].back_material
							sel=[]
							if e[1].is_a?(Sketchup::Group) || e[1].is_a?(Sketchup::ComponentInstance)
								sel=e[0].parent.entities.to_a if sels.include?(e[1])
							else
								sel=(e[0].parent.entities.to_a) & sels
							end
							sel.each{|e1|
								if e1.is_a?Sketchup::Drawingelement
									if e1.is_a?(Sketchup::Face) && !e[3] && (e1.back_material==material)
										back_materials<< [e1,e[2]]
									elsif (e1.material==material)
										materials << [e1,e[2]]
									end
								end
							}
						}
					elsif ctrl_press
						selects.each{|e|
							id =e[1].entityID.to_s + "." + e[0].entityID.to_s
							material=@picked_elements[id][3] ? @picked_elements[id][0].material : @picked_elements[id][0].back_material
							faces=[]
							faces=(e[0].all_connected.to_a).grep(Sketchup::Drawingelement) if e[0].is_a?(Sketchup::Face) || e[0].is_a?(Sketchup::Edge)
							sel=[]
							if e[1].is_a?(Sketchup::Group) || e[1].is_a?(Sketchup::ComponentInstance)
								sel=faces if sels.include?(e[1])
							else
								sel=faces & sels
							end
							sel.each{|e1|
								if e1.is_a?Sketchup::Drawingelement
									if e1.is_a?(Sketchup::Face) && !e[3] && (e1.back_material==material)
										back_materials<< [e1,e[2]]
									elsif (e1.material==material)
										materials << [e1,e[2]]
									end
								end
							}
						}
					else
						sels1=Sketchup.active_model.selection.to_a.grep(Sketchup::Drawingelement)
						selects.each{|e|
							if e[3]
								if sels1.empty?
									materials << [e[0],e[2]]
								else
									if sels1.include? e[1]
										materials <<  [e[0],e[2]] if e[0]!=e[1]
										sels1.each{|e1| materials << [e1,e[2]]}
									else
										materials <<  [e[0],e[2]]
									end
								end
							else
								if sels1.empty?
									back_materials<<  [e[0],e[2]]
								else
									if sels1.include? e[1]
										sels1.each{|e1|
											if e1.is_a?Sketchup::Face
												back_materials << [e1,e[2]]
											else
												materials <<  [e1,e[2]]
											end
										}
									else
										if e[0].is_a?Sketchup::Face
											back_materials <<  [e[0],e[2]]
										else
											materials <<  [e[0],e[2]]
										end
									end
								end
							end
						}
					end
					# paint_material(materials,back_materials)
				elsif paint_mode==1
					selects=@picked_elements.each_value.collect{|e| [e[0],e[1],e[2],e[3]]}.uniq{|e| e[1]}
					sels=Sketchup.active_model.selection.to_a.grep(Sketchup::Drawingelement)
					sels=Sketchup.active_model.active_entities.to_a if sels.empty?
					if @shift_press
						selects.each{|e|
							id =e[1].entityID.to_s + "." + e[0].entityID.to_s
							material=get_material id
							sels.each{|e1|
								if e1.is_a?Sketchup::Drawingelement
									back_materials << [e1,e[2]] if e1.is_a?(Sketchup::Face) && (e1.back_material==material)
									materials<< [e1,e[2]] if (e1.material==material)
								end
							}
						}
					elsif @ctrl_press
						selects.each{|e|
							faces=[]
							faces=e[0].all_connected.to_a.grep(Sketchup::Drawingelement) if e[0].is_a?(Sketchup::Face) || e[0].is_a?(Sketchup::Edge)
							id =e[1].entityID.to_s + "." + e[0].entityID.to_s
							faces<<e[1]
							material=get_material id
							(faces & sels).each{|e1|
								if e1.is_a?Sketchup::Drawingelement
									back_materials << [e1,e[2]] if e1.is_a?(Sketchup::Face) && (e1.back_material==material)
									materials<< [e1,e[2]] if (e1.material==material)
								end
							}
						}
					else
						sels1=Sketchup.active_model.selection.to_a.grep(Sketchup::Drawingelement)
						selects.each{|e|
							if e[3]
								if sels1.empty?
									materials << [e[1],e[2]]
								else
									if sels1.include? e[1]
										sels1.each{|e1| materials << [e1,e[2]]}
									else
										materials << [e[1],e[2]]
									end
								end
							else
								if sels1.empty?
									if e[1].is_a?Sketchup::Face
										back_materials << [e[1],e[2]]
									else
										materials << [e[1],e[2]]
									end
								else
									if sels1.include? e[1]
										sels1.each{|e1|
											if e1.is_a?Sketchup::Face
												back_materials << [e1,e[2]]
											else
												materials << [e1,e[2]]
											end
										}
									else
										if e[1].is_a?Sketchup::Face
											back_materials << [e[1],e[2]]
										else
											materials << [e[1],e[2]]
										end
									end
								end
							end
						}
					end
					# paint_material(materials,back_materials)
				end
				return [materials,back_materials]
			end
			def paint_face_soft(face_material_start=nil)
				model=Sketchup.active_model
				sel=model.selection
				@picked_elements.values.each{|p|
					materials=[]
					back_materials=[]
					if p[0].is_a?(Sketchup::Face) && ((p[0]==p[1] && @paint_mode==1) ||  @paint_mode!=1)
						picked_elements,faces_soft_hash ,faces_list= select_surface p[0],p[1],p[2],p[3]
						faces_soft=faces_soft_hash.values[0]
						if faces_soft && faces_soft[0]
							faces_soft.each_index{|i|
								fs=faces_soft[i][0].collect{|e| [e[0],e[2]]}
								if i>0
									fr=faces_soft[i][1]
									face_material=[fr[0],fr[3],fr[2]]
								else
									fr=faces_soft[0][1]
									if !face_material_start
										face_material=[fr[0],fr[3],fr[2]]
									else
										face_material=face_material_start
									end
								end
								if fr[3]
									materials << [fs,[],face_material]
								else
									back_materials << [[],fs,face_material]
								end
							}
							materials.each{|m| paint_material(m[0],m[1],m[2])}
							back_materials.each{|m| paint_material(m[0],m[1],m[2])}
						end
					end
				}
			end
			def paint_material(materials,back_materials,face_material=@face_material)
				materials.uniq!{|m| m[0]}
				back_materials.uniq!{|m| m[0]}
				(p "No material" ;return) if (materials+back_materials).empty?
				model=Sketchup.active_model
				edit_transform=model.edit_transform
				material=Sketchup.active_model.materials.current
				materials.uniq.each{|f| f[0].material = material if (f[0].is_a?Sketchup::Drawingelement) && !(f[0].is_a?Sketchup::Face)}
				faces=materials.select{|e| e if e[0].is_a? Sketchup::Face}
				points=nil
				if face_material && face_material[0].is_a?(Sketchup::Face) &&
					((face_material[0].material && face_material[0].material.texture && face_material[1]) ||
					(face_material[0].back_material && face_material[0].back_material.texture && !face_material[1]))
					face=face_material[0]
					vec=face_material[1]
					uvHelp = face.get_UVHelper(true, true)
					t= face_material[2]
					t=IDENTITY if !t
					origin=S4U::S4u_paint.get_min_max(face.outer_loop.vertices)[0]
					normal=(vec ? face.normal : face.normal.reverse)
					pts = []
					pts<< Geom::Point3d.new( [0, 0, 0] )
					pts<< Geom::Point3d.new( [100, 0, 0] )
					pts<< Geom::Point3d.new( [100, 100, 0] )
					pts<< Geom::Point3d.new( [0, 100, 0] )
					sc=S4U::S4u_paint.get_scale t
					points=[]
					if face.parent.entities==model.active_entities
						trans_source_in=Geom::Transformation.new origin.position.transform(edit_transform.inverse),normal.transform(edit_transform.inverse)
						ts=Geom::Transformation.scaling origin.position.transform(edit_transform.inverse),sc[0],sc[1],sc[2]
						pts1=pts.collect{|p| p.transform(edit_transform*ts.inverse*trans_source_in)}
						if vec
							points=pts1.collect{|e| uvHelp.get_front_UVQ(e)}
						else
							points=pts1.collect{|e| uvHelp.get_back_UVQ(e)}
						end
					else
						ts=Geom::Transformation.scaling origin.position,sc[0],sc[1],sc[2]
						trans_source_in=Geom::Transformation.new origin.position,normal
						pts1=pts.collect{|p| p.transform(ts.inverse*trans_source_in)}
						if vec
							points=pts1.collect{|e| uvHelp.get_front_UVQ(e)}
						else
							points=pts1.collect{|e| uvHelp.get_back_UVQ(e)}
						end
					end
					front=true
					if face
						faces.each{|f|
							trans=f[1]
							sc=S4U::S4u_paint.get_scale trans
							normal=f[0].normal
							origin=S4U::S4u_paint.get_min_max(f[0].outer_loop.vertices)[0]
							if f[0].parent.entities==model.active_entities
								ts1=Geom::Transformation.scaling origin.position.transform(edit_transform.inverse),sc[0],sc[1],sc[2]
								trans_face_in=Geom::Transformation.new origin.position.transform(edit_transform.inverse),normal.transform(edit_transform.inverse)
								trans1=S4U::S4u_paint.copy(trans_source_in,trans_face_in)
								pts2=pts.collect{|p| p.transform(edit_transform*ts.inverse*trans1)}
							else
								ts1=Geom::Transformation.scaling origin.position,sc[0],sc[1],sc[2]
								trans_face_in=Geom::Transformation.new origin.position,normal
								trans1=S4U::S4u_paint.copy(t*trans_source_in,trans*trans_face_in)
								pts2=pts.collect{|p| p.transform(trans.inverse*trans1)}
							end
							ps= pts2.zip points
							begin
							f[0].position_material(material, ps.flatten, front)
							rescue
							p "error maping"
							f[0].material=material
							end
						}
						front=false
						back_materials.uniq.each{|f|
							trans=f[1]
							sc=S4U::S4u_paint.get_scale trans
							normal=f[0].normal.reverse
							origin=S4U::S4u_paint.get_min_max(f[0].outer_loop.vertices)[0]
							if f[0].parent.entities==model.active_entities
								ts1=Geom::Transformation.scaling origin.position.transform(edit_transform.inverse),sc[0],sc[1],sc[2]
								trans_face_in=Geom::Transformation.new origin.position.transform(edit_transform.inverse),normal.transform(edit_transform.inverse)
								trans1=S4U::S4u_paint.copy(trans_source_in,trans_face_in)
								pts2=pts.collect{|p| p.transform(edit_transform*ts.inverse*trans1)}
							else
								ts1=Geom::Transformation.scaling origin.position,sc[0],sc[1],sc[2]
								trans_face_in=Geom::Transformation.new origin.position,normal
								trans1=S4U::S4u_paint.copy(t*trans_source_in,trans*trans_face_in)
								pts2=pts.collect{|p| p.transform(trans.inverse*trans1)}
							end
							ps= pts2.zip points
							begin
							f[0].position_material(material, ps.flatten, front)
							rescue
							p "error maping back"
							f[0].material=material
							end
						}
					end
				elsif material && material.texture
					pts = []
					x,y = material.texture.width,material.texture.height
					pts[0] = Geom::Point3d.new [0, 0, 0]
					pts[1] = Geom::Point3d.new [x, 0, 0]
					pts[2] = Geom::Point3d.new [x, y, 0]
					pts[3] = Geom::Point3d.new [0, y, 0]
					pts1 = []
					x,y =1,1
					pts1[0] = Geom::Point3d.new [0, 0, 0]
					pts1[1] = Geom::Point3d.new [x, 0, 0]
					pts1[2] = Geom::Point3d.new [x, y, 0]
					pts1[3] = Geom::Point3d.new [0, y, 0]
					front=true
					sc=S4U::S4u_paint.get_scale edit_transform
					ts=Geom::Transformation.scaling ORIGIN,sc[0],sc[1],sc[2]
					faces.each{|f|
						trans=f[1]
						sc=S4U::S4u_paint.get_scale trans
						ts1=Geom::Transformation.scaling ORIGIN,sc[0],sc[1],sc[2]
						normal=f[0].normal
						if f[0].parent.entities==model.active_entities
							trans_face_in=Geom::Transformation.new ORIGIN,normal.transform(edit_transform.inverse)
							pts2=pts.collect{|p| p.transform(edit_transform*ts.inverse*trans_face_in)}
						else
							trans_face_in=Geom::Transformation.new ORIGIN,normal
							pts2=pts.collect{|p| p.transform(ts1.inverse*trans_face_in)}
						end
						ps= pts2.zip pts1
						begin
						f[0].position_material(material,ps.flatten,front)
						rescue
						p "error maping"
						f[0].material=material
						end
					}
					front=false
					back_materials.each{|f|
						trans=f[1]
						sc=S4U::S4u_paint.get_scale trans
						ts1=Geom::Transformation.scaling ORIGIN,sc[0],sc[1],sc[2]
						normal=f[0].normal.reverse
						if f[0].parent.entities==model.active_entities
							trans_face_in=Geom::Transformation.new ORIGIN,normal.transform(edit_transform.inverse)
							pts2=pts.collect{|p| p.transform(edit_transform*ts.inverse*trans_face_in)}
						else
							trans_face_in=Geom::Transformation.new ORIGIN,normal
							pts2=pts.collect{|p| p.transform(ts1.inverse*trans_face_in)}
						end
						ps= pts2.zip pts1
						begin
						f[0].position_material(material,ps.flatten,front)
						rescue
						p "error maping back"
						f[0].material=material
						end
					}
				else
					faces.each{|f|  f[0].material = material }
					back_materials.each{|f| f[0].back_material = material}
				end
			end
			def align_material(materials,back_materials,picked_edge=@picked_edge)
				materials=materials.select{|f| f if  f[0].valid? && f[0].is_a?(Sketchup::Face) && f[0].material && f[0].material.texture}
				back_materials=back_materials.select{|f| f if  f[0].valid? && f[0].is_a?(Sketchup::Face) && f[0].back_material && f[0].
				back_material.texture}
				if (materials+back_materials).empty?
					UI.messagebox @strings["The Selected Face is No Material"]
					return
				end
				return if !picked_edge
				pts_map = []
				x,y =1,1
				pts_map[0] = Geom::Point3d.new [0, 0, 0]
				pts_map[1] = Geom::Point3d.new [x, 0, 0]
				pts_map[2] = Geom::Point3d.new [x, y, 0]
				pts_map[3] = Geom::Point3d.new [0, y, 0]
				front=true
				materials.each{|f|
					material=f[0].material
					trans=f[1]
					normal=f[0].normal
					mx,my = material.texture.width,material.texture.height
					sc=S4U::S4u_paint.get_scale trans
					ts=Geom::Transformation.scaling ORIGIN,sc[0],sc[1],sc[2]
					#ts=trans
					ed=picked_edge.collect{|p| p.transform(trans.inverse)}
					min,max=S4U::S4u_paint.get_min_max(ed)
					min=min.project_to_plane(f[0].plane)
					max=max.project_to_plane(f[0].plane)
					line=max-min
					vz=normal
					vx=line
					vy=vz.cross vx
					if vy.length!=0
						trans_face_in= Geom::Transformation.new(vx,vy,vz,min)
						#ang,sx,sy=S4U::S4u_paint.get_info_texture(f[0],trans,front)
						x,y=mx,my
						pts = []
						pts[0] = Geom::Point3d.new [0, 0, 0]
						pts[1] = Geom::Point3d.new [x, 0, 0]
						pts[2] = Geom::Point3d.new [x, y, 0]
						pts[3] = Geom::Point3d.new [0, y, 0]
						pts2=pts.collect{|p| p.transform(ts.inverse*trans_face_in)}
						ps= pts2.zip pts_map
						begin
						f[0].position_material(material,ps.flatten,front)
						rescue
						p "error maping"
						f[0].material=material
						end
					end
				}
				front=false
				back_materials.each{|f|
					material=f[0].back_material
					trans=f[1]
					normal=f[0].normal.reverse
					mx,my = material.texture.width,material.texture.height
					sc=S4U::S4u_paint.get_scale trans
					ed=picked_edge.collect{|p| p.transform(trans.inverse)}
					min,max=S4U::S4u_paint.get_min_max(ed)
					min=min.project_to_plane(f[0].plane)
					max=max.project_to_plane(f[0].plane)
					line=max-min
					vz=normal
					vx=line
					vy=vz.cross vx
					if vy.length!=0
						trans_face_in= Geom::Transformation.new(vx,vy,vz,min)
						# ang,sx,sy=S4U::S4u_paint.get_info_texture(f[0],trans,front)
						# x,y =sc[0]*sx*mx,sc[1]*sy*my
						x,y=mx,my
						pts = []
						pts[0] = Geom::Point3d.new [0, 0, 0]
						pts[1] = Geom::Point3d.new [x, 0, 0]
						pts[2] = Geom::Point3d.new [x, y, 0]
						pts[3] = Geom::Point3d.new [0, y, 0]

						pts2=pts.collect{|p| p.transform(trans_face_in)}
						ps= pts2.zip pts_map
						begin
						f[0].position_material(material,ps.flatten,front)
						rescue
						p "error maping back"
						f[0].material=material
						end
					end
				}
			end
			def rotate_material(materials,back_materials,angle=@angle,origin=nil)
				materials=materials.select{|f| f if  f[0].valid? && f[0].is_a?(Sketchup::Face) && f[0].material && f[0].material.texture}
				back_materials=back_materials.select{|f| f if  f[0].valid? && f[0].is_a?(Sketchup::Face) && f[0].back_material && f[0].
				back_material.texture}
				if (materials+back_materials).empty?
					UI.messagebox @strings["The Selected Face is No Material"]
					return
				end
				pts1 = []
				x,y =1,1
				pts1[0] = Geom::Point3d.new [0, 0, 0]
				pts1[1] = Geom::Point3d.new [x, 0, 0]
				pts1[2] = Geom::Point3d.new [x, y, 0]
				pts1[3] = Geom::Point3d.new [0, y, 0]
				front=true
				materials.each{|f|
					material=f[0].material
					normal=f[0].normal
					x,y = material.texture.width,material.texture.height
					ratio=y.to_f/x
					trans=f[1]
					sc=S4U::S4u_paint.get_scale trans
					ts=Geom::Transformation.scaling ORIGIN,sc[0],sc[1],sc[2]
					if @origin && @origin.valid?
						origin=@origin.position.transform(trans.inverse)
						origin=origin.project_to_plane f[0].plane
					else
						origin=S4U::S4u_paint.get_min_max(f[0].outer_loop.vertices)[0]
					end
					trans_face_in=Geom::Transformation.new origin,normal
					pts2=pts1.collect{|p| p.transform(ts.inverse*trans_face_in)}

					uvHelp=f[0].get_UVHelper(front,!front)
					points=pts2.collect{|p| uvHelp.get_front_UVQ(p)}
					tr=Geom::Transformation.rotation origin,normal,@angle
					pts3=pts1.collect{|p| p.transform(ts.inverse*trans_face_in)}
					pts3=pts3.collect{|p| p.transform(tr)}
					ps= pts3.zip points
					f[0].position_material(material,ps.flatten,front)

				}
				front=false
				back_materials.each{|f|
					material=f[0].back_material
					normal=f[0].normal.reverse
					x,y = material.texture.width,material.texture.height
					ratio=y.to_f/x
					trans=f[1]
					sc=S4U::S4u_paint.get_scale trans
					ts=Geom::Transformation.scaling ORIGIN,sc[0],sc[1],sc[2]
					origin=S4U::S4u_paint.get_min_max(f[0].outer_loop.vertices)[0]
					trans_face_in=Geom::Transformation.new origin,normal
					pts2=pts1.collect{|p| p.transform(ts.inverse*trans_face_in)}

					uvHelp=f[0].get_UVHelper(front,!front)
					points=pts2.collect{|p| uvHelp.get_back_UVQ(p)}
					tr=Geom::Transformation.rotation origin,normal,@angle
					pts3=pts1.collect{|p| p.transform(ts.inverse*trans_face_in)}
					pts3=pts3.collect{|p| p.transform(tr)}
					ps= pts3.zip points
					f[0].position_material(material,ps.flatten,front)

				}
			end
			def move_material(materials,back_materials,picked_edge)
				materials=materials.select{|f| f if f[0].valid? && f[0].is_a?(Sketchup::Face) && f[0].material && f[0].material.texture}
				back_materials=back_materials.select{|f| f if f[0].valid? && f[0].is_a?(Sketchup::Face) && f[0].back_material && f[0].
				back_material.texture}
				if (materials+back_materials).empty?
					# p @strings["The Selected Face is No Material"]
					return
				end
				pts1 = []
				x,y =1,1
				pts1[0] = Geom::Point3d.new [0, 0, 0]
				pts1[1] = Geom::Point3d.new [x, 0, 0]
				pts1[2] = Geom::Point3d.new [x, y, 0]
				pts1[3] = Geom::Point3d.new [0, y, 0]
				front=true
				materials.each{|f|
					material=f[0].material
					normal=f[0].normal
					trans=f[1]
					trans_face_in=Geom::Transformation.new ORIGIN,normal
					pts2=pts1.collect{|p| p.transform(trans.inverse*trans_face_in)}
					uvHelp=f[0].get_UVHelper(front,!front)
					points=pts2.collect{|p| uvHelp.get_front_UVQ(p)}
					min,max=picked_edge.collect{|p| p.transform(trans.inverse)}.collect{|p| p.project_to_plane(f[0].plane)}
					tm=Geom::Transformation.translation(max-min)
					pts3=pts1.collect{|p| p.transform(tm*trans.inverse*trans_face_in)}
					ps= pts3.zip points
					f[0].position_material(material,ps.flatten,front)
				}
				front=false
				back_materials.each{|f|
					material=f[0].back_material
					normal=f[0].normal.reverse
					trans=f[1]
					trans_face_in=Geom::Transformation.new ORIGIN,normal
					pts2=pts1.collect{|p| p.transform(trans.inverse*trans_face_in)}
					uvHelp=f[0].get_UVHelper(front,!front)
					points=pts2.collect{|p| uvHelp.get_back_UVQ(p)}
					min,max=picked_edge.collect{|p| p.transform(trans.inverse)}.collect{|p| p.project_to_plane(f[0].plane)}
					tm=Geom::Transformation.translation(max-min)
					pts3=pts1.collect{|p| p.transform(tm*trans.inverse*trans_face_in)}
					ps= pts3.zip points
					f[0].position_material(material,ps.flatten,front)
				}
			end
			def delete_material
				model=Sketchup.active_model
				sel=model.selection.to_a
				objects=[]
				back_materials=[]
				materials=[]
				selects=@picked_elements.each_value.collect{|e| [e[0],e[3],e[1]] }.uniq{|e| e[2]}
				sels=Sketchup.active_model.selection.to_a.grep(Sketchup::Drawingelement)
				sels=Sketchup.active_model.active_entities.to_a if sels.empty?
				faces_soft=[]
				if @faces_soft && @faces_soft.length>0
					@faces_soft.each{|f| faces_soft+=(f.collect{|f1| f1[0]}).flatten(1).collect{|f2| f2[0]}.flatten(1)}
					faces_soft.flatten!
					faces_soft.uniq!
				end
				sels+=faces_soft
				if @shift_press
					selects.each{|e|
						id =e[2].entityID.to_s + "." + e[0].entityID.to_s
						material=get_material id
						sels.each{|e1|
							if e1.is_a?Sketchup::Drawingelement
								back_materials << e1 if e1.is_a?(Sketchup::Face) #&& (e1.back_material==material)
								materials<< e1 #if (e1.material==material)
							end
						}
					}
				elsif @ctrl_press
					selects.each{|e|
						faces=e[0].all_connected.to_a.grep(Sketchup::Drawingelement)
						id =e[2].entityID.to_s + "." + e[0].entityID.to_s
						faces<<e[2]
						material=get_material id
						(faces & sels).each{|e1|
							if e1.is_a?Sketchup::Drawingelement
								back_materials << e1 if e1.is_a?(Sketchup::Face) #&& (e1.back_material==material)
								materials<< e1 #if (e1.material==material)
							end
						}
					}
				else
					sels1=Sketchup.active_model.selection.to_a.grep(Sketchup::Drawingelement)
					sels1+=faces_soft
					selects.each{|e|
						if e[1]
							if sels1.empty?
								materials<<e[2]
							else
								sels1.each{|e1| materials << e1}
							end
						else
							if sels1.empty?
								back_materials<<e[2]
							else
								sels1.each{|e1|
									if e1.is_a?Sketchup::Face
										back_materials << e1
									else
										materials << e1
									end
								}
							end
						end
					}
				end
				objects = materials + back_materials
				S4U::S4u_paint.set_materials(objects) if !objects.empty?
				if !Sketchup.active_model.rendering_options["DrawHidden"]
					mats=get_face_soft.flatten!
					mats.each{|f| f.material=nil ; f.back_material=nil} if mats && !mats.empty?
				end
			end
			def get_face_soft
				model=Sketchup.active_model
				materials=[]
				back_materials=[]
				@picked_elements.values.each{|p|
					if p[0].is_a?(Sketchup::Face) && ((p[0]==p[1] && @paint_mode==1) ||  @paint_mode!=1)
						faces=S4U::S4u_paint.find_connected(p[0])
						if p[3]
							materials=faces
						else
							back_materials=faces
						end
					end
				}
				return [materials,back_materials]
			end
			def replace_material
				model=Sketchup.active_model
				sel=model.selection.to_a
				sel=model.active_entities.to_a if sel.empty?
				material=Sketchup.active_model.materials.current
				id = @ph_select.best_picked.entityID.to_s + "." + @ph_select.picked_face.entityID.to_s
				material1=get_material(id)
				edit_material= material1 ? material1 : @edit_material
				objects=sel.select{|e| e.is_a?Sketchup::Drawingelement}
				if objects.length>0
					selects=[]
					objects1=[]
					faces=[]
					objects.each{|e|
						if (e.is_a?Sketchup::Group) || (e.is_a?Sketchup::ComponentInstance)
							faces+=S4U::S4u_paint.select_faces(e)
							objects1+=S4U::S4u_paint.select_elements(e)
						else
							selects<<e
						end
					}
					selects+=faces.transpose[0].uniq if faces && faces.length>0
					selects.each{|f|
							f.material = material if f.material==material1
							f.back_material=material if (f.is_a?Sketchup::Face) && (f.back_material==material1)
					}
					objects.uniq.each{|e| e.material=material if e.material==material1 || (material1==edit_material && !e.material && !edit_material)
								}
					objects1.uniq.each{|e| e.material=material if e.material==material1 }
				end
			end
			def get_material(select)
				material=nil
				f=@picked_elements[select][0]
				if f.is_a?Sketchup::Face
					material = @picked_elements[select][3] ? f.material : f.back_material
				end
				if !material
					@ph_select.path_at(@picked_elements[select][4]).to_a.reverse[1..-1].each{|e|
						(material=e.material;break) if e.material
					}
				end
				return material
			end
			def onKeyDown(key, repeat, flags, view)
				if ((key==VK_SHIFT) && IS_WIN ) || ((key==VK_SHIFT) && IS_OSX)
					@shift_press=true
					onSetCursor
					self.setstatus(view)
				elsif ((key==VK_CONTROL) && IS_WIN ) || ((key==VK_ALT) && IS_OSX)
					@ctrl_press=true
					onSetCursor
					self.setstatus(view)
				elsif ((key==VK_ALT) && IS_WIN ) || ((key==VK_CONTROL) && IS_OSX)
					@alt_press=true
					onSetCursor
					self.setstatus(view)
					return true
				end
			end
			def onKeyUp(key, repeat, flags, view)
				if ((key==VK_SHIFT) && IS_WIN ) || ((key==VK_SHIFT) && IS_OSX)
					@shift_press=nil
					onSetCursor
					self.setstatus(view)
				elsif ((key==VK_CONTROL) && IS_WIN ) || ((key==VK_ALT) && IS_OSX)
					@ctrl_press=nil
					onSetCursor
					self.setstatus(view)
				elsif ((key==VK_ALT) && IS_WIN ) || ((key==VK_CONTROL) && IS_OSX)
					@alt_press=false
					onSetCursor
					self.setstatus(view)
				end
			end
			def onUserText(text, view)
				if @paint_mode==4
					input=text.to_f
					@angle=input.degrees if input>-180 && input<=180
				else
					input=text.to_i
					@aperture=input if input>=0 && input<=50
				end
				self.setstatus
			end
			def onCancel(reason, view)
				self.reset(view)
			end
			def reset(view=nil)
				@picked_elements.clear
				@ip.clear
				@faces_soft_hash={}
				@faces_soft=[]
				@ctrl_press=nil
				@alt_press=nil
				@shift_press=nil
				@picked_edge=nil
				@point_start.clear
				@point_end.clear
				@x,@y=0,0
				self.setstatus(view)
			end
			def setstatus(view=nil,mes=nil)
				text=@paint_list[@paint_mode]
				if @paint_mode==0 || @paint_mode==1
					text+= " | "
					if @paint_mode==0
						text+="Select Face To Paint" + " | "
					elsif @paint_mode==1
						text+="Select Object To Paint" + " | "
					end
					if IS_OSX
						text+=@strings.GetString("Hold Shift -> All Faces In Model | Hold Alt -> All Connected | Hold Ctrl -> Sample Material")
					else
						text+=@strings.GetString("Hold Shift -> All Faces In Model | Hold Ctrl -> All Connected | Hold Alt -> Sample Material")
					end
				elsif @paint_mode==4
					text+=" | "
					if IS_OSX
						text+=@strings.GetString("Click Face to Rotate,Hold Alt -> Set Origin | Hold Shift -> Move Material | Hold Ctrl -> Select Line to Align")
					else
						text+=@strings.GetString("Click Face to Rotate,Hold Ctrl -> Set Origin | Hold Shift -> Move Material | Hold Alt -> Select Line to Align")
					end
				end
				if mes
					text+= " | " + mes
				else
					tooltip=@strings["Current Material:"]+" "
					material=Sketchup.active_model.materials.current
					tooltip+= material ? material.display_name : @strings["Default"]
					text+= " | " + tooltip
				end
				Sketchup.status_text=text
				if @paint_mode==4
					Sketchup.vcb_label = @strings["Angle(-180<a<=180)"]
					Sketchup.vcb_value = @angle.radians.round(2).to_s
				else
					Sketchup.vcb_label = @strings["Aperture(0-50)"]
					Sketchup.vcb_value = @aperture.to_s
				end
				view.invalidate if view
			end
			def resume(view)
				self.setstatus(view)
			end
		end
		def self.paint(mode)
			@setting=[0,2,Math::PI*0.5,false] if !@setting
			@mode=0 if !@mode
			@setting[0]=mode if mode<7
			Sketchup.active_model.select_tool Paint.new(@setting)
		end
		def self.main(mode)
			self.paint(mode)
		end
	end
end